program ANacoustic
  ! Program to calculate exact solutions to acoustic wave equation in 2d
  
  use sep
  use fftw_mod
  use fftshift_io_mod
  implicit none
  
  integer     :: n1,n2,nt,nw
  real        :: d1,d2,dt,dw,x
  real        :: o1,o2,ot,ow

  integer     :: iw,it,ii,jj,kk,js

  ! Source and receiver positions variables
  integer,allocatable               :: sources_zxt   (:,:),receiver_zxt   (:,:)
  real,allocatable                  :: sources_zxt_in(:,:),receiver_zxt_in(:,:)
  integer                           :: ns,nsx,nsz,fsx,fsz,jsx,jsz,stype
  integer                           :: nr,nrt,nrx,nrz,frx,frz,jrx,jrz,rtype
  integer                           :: isx,isz,irx,irz,is,ir,is2,ns2,ir2,nr2
  
  real   , allocatable :: data  (:,:,:)
  complex, allocatable :: dataw (:,:,:)
  complex, allocatable :: source(:,:)
  real   , allocatable ::     w (:)
  
  real              :: model(3)
  real              :: rin,kin,vin
  
  logical     :: sym_t
  real        :: f0,w0,t0,fact

  real,parameter :: pi = 3.14159265

  real        :: WEIGHT,rr,smth_fact
  integer     :: sx,sz,ism,jsm,smth,sind
  
  
  ! Header IO
  call sep_init()
  
  ! Geometry
  call from_param('nt',nt,512)
  call from_param('n1',n1,500)
  call from_param('n2',n2,500)
  
  call from_param('dt',dt,.001)
  call from_param('d1',d1,  1.)
  call from_param('d2',d2,  1.)
  
  call from_param('ot',ot,  0.)
  call from_param('o1',o1,  0.)
  call from_param('o2',o2,  0.)
  
  call from_param('sym_t',sym_t,.false.)
  if (sym_t) then;  call to_history("n1",2*nt);
  else;             call to_history("n1",  nt); end if
  call to_history("d1",dt);
  if (sym_t) then;  call to_history("o1",-nt*dt);
  else;             call to_history("o1",    ot); end if
  
  call to_history('n2',n1)
  call to_history('n3',n2)

  call to_history('d2',d1)
  call to_history('d3',d2)
  
  call to_history('o2',o1)
  call to_history('o3',o2)

  call to_history("label1",'time (s)')
  call to_history("label2",'z position (m)')
  call to_history("label3",'x position (m)')
  
  ! Model
  call from_param('r',model(2),0.);
  call from_param('k',model(3),0.);
  if ( rin.ne.0. .and. kin.ne.0. ) then
    model(1) = 1./sqrt(dble(model(2))*dble(model(3)))
  else
    call from_param('v',model(1),0.);
  end if
  ! Make model complete and or self consistent
  if      (model(2).ne.0.) then
    model(3) = dble(model(2))/(dble(model(1))**2.)
  else if (model(3).ne.0) then
    model(2) = dble(model(3))/(dble(model(1))**2.)
  end if
  
  ! *********************************************************************
  !  Source/Receicer Section
  ! *********************************************************************
  
  ! Receiver Output
  call from_param("nrt",nrt,-1)
  call from_param("nrz",nrz,1 )
  call from_param("nrx",nrx,1 )
  call from_param("frz",frz,n1/2+1)
  call from_param("frx",frx,n2/2+1)
  call from_param("jrz",jrz,1 )
  call from_param("jrx",jrx,1 )
  call from_param("rtype",rtype,1)
  nr = nrz*nrx
  if (exist_file("receivers_zxt")) then
    call from_aux("receivers_zxt","n2",nr,nr)
  end if
  allocate(receiver_zxt(nr,4));  receiver_zxt = 0;
  if (exist_file("receiver_zxt")) then
    allocate(receiver_zxt_in(nr,4)); receiver_zxt_in=0.
    call sreed("receiver_zxt",receiver_zxt_in,4*size(receiver_zxt_in))
    do ir=1,nr
      receiver_zxt(ir,:) = (/ 1+nint((receiver_zxt_in(ir,1)-o1)/d1), &
                              1+nint((receiver_zxt_in(ir,2)-o2)/d2), &
                              1+nint( receiver_zxt_in(ir,3)/dt), int(receiver_zxt_in(ir,4)) /)
    end do
    deallocate(receiver_zxt_in)
  else
    do irx=1,nrx
    do irz=1,nrz
      receiver_zxt((irx-1)*nrz+irz,:) = (/ 1+frz+(irz-1)*jrz,1+frx+(irx-1)*jrx,1,rtype /)
    end do
    end do
  end if

  ! Sources Input grid
  call from_param("nsz",nsz,1 )
  call from_param("nsx",nsx,1 )
  call from_param("fsz",fsz,n1/2+1)
  call from_param("fsx",fsx,n2/2+1)
  call from_param("jsz",jsz,1 )
  call from_param("jsx",jsx,1 )
  call from_param("stype",stype,1)
  call from_param("t0" ,t0 ,0.)
  ns = nsz*nsx
  if (exist_file("sources_zxt")) then
    call from_aux("sources_zxt","n1",ns)
    allocate(sources_zxt   (ns,4)); sources_zxt = 0.
    allocate(sources_zxt_in(ns,4)); sources_zxt_in = 0.
    call sreed("sources_zxt",sources_zxt_in,4*size(sources_zxt_in))
    do is=1,ns
      sources_zxt(is,:) = (/ 1+nint((sources_zxt_in(is,1)-o1)/d1), &
                             1+nint((sources_zxt_in(is,2)-o2)/d2), &
                             1+nint( sources_zxt_in(is,3)/dt), int(sources_zxt_in(is,4)) /)
    end do
  else
    allocate(sources_zxt   (ns,4)); sources_zxt = 0.
    do isx=1,nsx
    do isz=1,nsz
      sources_zxt((isx-1)*nsz+isz,:) = (/ 1+fsz+(isz-1)*jsz,1+fsx+(isx-1)*jsx,1,stype /)
    end do
    end do
  end if

  ! Source
  call from_param("f0",f0,25.); w0 = 2.*pi*f0

  call from_param("nsz",nsz,   1)
  call from_param("nsx",nsx,   1)
  call from_param("fsz",fsz,n1/2+1)
  call from_param("fsx",fsx,n2/2+1)
  call from_param("jsz",jsz,   1)
  call from_param("jsx",jsx,   1)

  call from_param("stype",stype,0)
  call from_param("smth",smth,0) ! Source smoothing over smth gridpoints
  smth_fact = 0.   ! Weighting factor because of source smoothing
  do jsm=-smth,smth
  do ism=-smth,smth
      rr        = sqrt(float((ism)**2+(jsm)**2))
      smth_fact = smth_fact + .5*(cos(rr/3.*pi)+1.)
  end do
  end do

  call sep_close()
  
    
  ! Calculate frequency units 
  nw =        nt+1
  dw = 2.*pi/(nt*2*dt)
  ow = 0.
  fact = 2./sqrt(real(2*nt))/dt  ! Amplitude correction factor
  
  ! Compute number of sources (including smoothed injection)
  sind=0
  do is=1,ns
    do jsm=-smth,smth
    do ism=-smth,smth
      sind=sind+1
    end do
    end do
  end do


  ! Allocate data-space and source-vector
  allocate( data (2*nt,n1,n2)); data   = 0.
  allocate( dataw(  nw,n1,n2)); dataw  = 0.
  allocate( source( nw, sind)); source = 0.
  allocate(      w( nw      ));      w = 0;
  
  ! Init IFFT
  call fftw1d_1_r2c_lop_init(nt_in=2*nt)
  
  ! Set up frequency vector
  do iw=1,nw
    w(iw) = ow+(iw-1)*dw
  end do
  w(1) = epsilon(w);
  
  ! Construct frequency domain wavelet
  do iw = 1,nw
    source(iw,:) = ( (2.*w(iw)**2.) / (sqrt(pi)*w0**3.) ) * exp(-w(iw)**2./w0**2.)
  end do
  
  ! Add common part of Green's function
  do iw=1,nw
    source(iw,:) =  1./sqrt(8.*pi*w(iw)/model(1))*source(iw,:)*cexp(cmplx(0.,-pi/4.))
  end do

  ! Convert to source type (and add smoothing weight)
  sind=0
  do is=1,ns
    
    do jsm=-smth,smth
    do ism=-smth,smth
    
    sind = sind + 1
    rr   = sqrt(float(ism**2+jsm**2))
    
    select case (stype)
      
      case (0)
        do iw=1,nw
          source(iw,sind) =                          source(iw,sind)*.5*(cos(rr/3.*pi)+1.)/smth_fact
        end do
      
      case (1)
        do iw=1,nw
          source(iw,sind) = model(2)*cmplx(0.,w(iw))*source(iw,sind)*.5*(cos(rr/3.*pi)+1.)/smth_fact
        end do
      
    end select
    
    
    end do
    end do
    
  end do

  ! Delay the sources appropriately

  ! Convert to source type (and add smoothing weight)
  sind=0
  do is=1,ns
    
    do jsm=-smth,smth
    do ism=-smth,smth
    
      sind = sind + 1

      if (allocated(sources_zxt_in)) t0=sources_zxt_in(is,3)

      if (t0.lt.nt*dt) then
        do iw=1,nw
          source(iw,sind) = source(iw,sind)*cexp(cmplx(0.,-w(iw)*t0))
        end do
      else
        do iw=1,nw
          source(iw,sind) = 0.
        end do
      end if

    end do
    end do
    
  end do


  ! Normalize frequency by velocity
  w = w/model(1)   
  
  ! Loop over grid points
  do jj=1,n2
  do ii=1,n1

  ! Loop over sources
  sind = 0
  do is=1,ns
  
    sx = fsx+(js-1)*jsx
    sz = fsz+(is-1)*jsz
    if (allocated(sources_zxt_in)) sx = sources_zxt(is,2)
    if (allocated(sources_zxt_in)) sz = sources_zxt(is,1)

    do jsm=sx-smth,sx+smth
    do ism=sz-smth,sz+smth
      
      sind = sind + 1

      ! Propagation distance
      x  = sqrt(       ((ii-ism)*d1)**2. + ((jj-jsm)*d2)**2.  )
      ! Propagate each frequency from source location to grid point
      do iw=1,nw
        dataw(iw,ii,jj) = dataw(iw,ii,jj) + 1./sqrt(x)*cexp(cmplx(0.,-w(iw)*x))*source(iw,sind)
      end do
    
    end do
    end do
  
  end do

  end do
  end do

  ! Fix singularity at source points
  sind=0.
  do is=1,ns
    sx = fsx+(js-1)*jsx
    sz = fsz+(is-1)*jsz
    if (allocated(sources_zxt_in)) sx = sources_zxt(is,2)
    if (allocated(sources_zxt_in)) sz = sources_zxt(is,1)
    do jsm=sx-smth,sx+smth
    do ism=sz-smth,sz+smth
      sind = sind + 1
      dataw(:,ism,jsm) = cexp(cmplx(0.,-pi/4.))*source(iw,sind)
    end do
    end do
  end do
  
  ! Do IFFT
  call fftw1d_1_r2c_lop(.true.,.false.,data,dataw)
  data = fact*data  ! Amplitude correction factor

  ! Write data
  do ii=1,n2
    if (sym_t) then;  call fftshift1d_srite("out" ,data( :  ,:,ii), 2*nt,n1,4);
    else;             call            srite("out", data(1:nt,:,ii),   nt*n1*4);  end if
  end do

  ! Deallocate and close
  call fftw1d_1_r2c_lop_close()
  deallocate( data );
  deallocate( dataw );
  deallocate( source );
  deallocate( w    );
  deallocate(sources_zxt)
  deallocate(receiver_zxt)
  if (allocated( sources_zxt_in)) deallocate( sources_zxt_in)
  if (allocated(receiver_zxt_in)) deallocate(receiver_zxt_in)
  call exit(0);

end program


