!	   		        Acoustic2d
!
!            Heterogenous acoustic modeling with constant density in 2d.
!
! 	solves:	   d^2 U / dt^2 = C^2(x,z) [d^2 U / dx^2 + d^2 U / dz^2 ]
!
!            by time stepping using an explicit finite-difference 
!	     solution to the acoustic wave equation accurate to 2nd order 
!            in time and 16th order in x and z.
!
!		Author: John Etgen, modified by Sjoerd de Ridder 
! 
! Example:
!
! Acoustic2d par=parfile <Hvel>Hdata snap=Hsnap
!
! This program is useful for producing shot records (1 or more) over an 
! arbitrary two-dimensional model of the subsurface. The input consists of a 
! 2-D model of P-wave velocities (z or depth being the "fast direction"), a 
! wavelet time function to use as the source, and a set of parameters that 
! describes the acquisition parameters of the synthetic survey.
! 
! The output of the program is a series (1 or more) shot records, the user
! specifies the number of time samples per trace, the number of traces,
! and the number of shots. Also the program will output snapshots of the
! 2-D wavefield at various time levels by specifying a snaphot interval
! in the parameters supplied to the program.
!
!	Necessary input files:
!
!		Hvel is the Header for the velocity model, stdin
!
!		parfile  contains the geometry parameters
!			 for the synthetic seismic survey
!
!	Output files:
!
!		Hdata    Header file for the output seismograms
!
!		Hsnap	      Header for the (optional) output snapshots
!
! **Necessary parameters****************************************************
! ****** most quantitites like distances refer to grid points **************
!
!	From Hvel:	n1 	# of gridpoints in z (becomes nz)
!			n2	# of gridpoints in x (becomes nx)
!			d1	z gridpoint spacing (becomes dz)
!			d2	x gridpoint spacing (becomes dx)
!
!	From par:	T	                Time (runtime)
!			nshots (default=1)       Number of shots to model
!			shot_ix_1 (default=nx/2)	Grid location of first shot
!			shot_inc (default=1)    Shot increment in gridpoints
!			trace_1	(default=1)     Location of first receiver
!					        for first shot
!		        ntraces (default=nx)     Number of receivers per shot 
!			shot_depth (default=1)  Shot depth (grid points)
!  			receiver_depth (def=1)  Geophone depth (in grid points)
!     			snap_i     		time steps between snapshots
!						(default=no snapshots) 
!                       rollalong (default=0)   roll the receiver spread along 
!						with shots? (1=yes 0=default=no)
!	From par:	fmin,fmax		Maximum frequency in source   
!
!
!			
!% end of self-doc
!                   		    
!                                   
!
!


program acoustic2d_new
  use sep
  use sources_mod
  use sep_timers_mod

  implicit none
  
  ! Grid parameters
  integer                           :: nx,nz,npad
  real                              :: x0,z0
  integer                           :: it,iz,ix,iy
  real                              :: T
  
  ! Computational time variables
  integer                           :: nt_comp,jt_verb
  real                              :: dz,dx,dt_comp
  real                              :: vmin,vmax

  ! Finite Difference variables
  real,dimension(:,:),allocatable   :: mult
  real,dimension(:,:),allocatable   :: a1x,a2x,a3x,a4x,a5x,a0,a1z,a2z,a3z,a4z,a5z
  real,allocatable                  :: u(:,:,:)
  real,allocatable                  :: data(:,:)

  ! Medium variables
  real,allocatable                  :: vv(:,:),rh(:,:)
  real                              :: rhoc

  ! Source and receiver positions variables
  integer,allocatable               :: sources_zxt(:,:),receiver_zxt(:,:)
  real,allocatable                  :: sources_zxt_in(:,:),receiver_zxt_in(:,:),vprev(:)
  integer                           :: ns,nsx,nsz,fsx,fsz,jsx,jsz,stype
  integer                           :: nr,nrt,nrx,nrz,frx,frz,jrx,jrz,rtype
  integer                           :: isx,isz,irx,irz,is,ir
  integer                           :: jt_data,nt_data
  real                              :: dt_data

  ! For snap shots
  real,allocatable                  :: usnap(:,:)
  integer                           :: jt_snap,nt_snap
  real                              :: dt_snap

  ! Source parameters
  real,allocatable                  :: sources(:,:)
  real                              :: f0,w0,smth_fact
  integer                           :: nst,smth

  ! ???
  real                              :: id,in
  real                              :: dxi,dzi
  real                              :: dt2,rr,delp
  integer                           :: ixe,ize,jt

  real,parameter                    :: pi=3.1415926535

  logical                           :: snap, ssquare, freesurf

  ! For timers
  logical           :: logic
  integer           :: tm1,tm2,tm3

  write(0,*) '----------------------------'
  write(0,*) 'Constant Density Acoustic FD'

  logic = init_sep_timers()
  logic = setup_next_timer("Total----",tm1)
  logic = setup_next_timer("Overhead-",tm2)
  logic = setup_next_timer("Core-----",tm3)
  call start_timer_num(tm1)
  call start_timer_num(tm2)
  
  call sep_init()
  ! *********************************************************************
  !  Setup Computational Grid
  ! *********************************************************************
  write(0,*) '----------------------------'
  write(0,*) 'Setup Computational Grid'
  
  ! Geometry of the experiments
  call from_history("n1",nz)
  call from_history("n2",nx)
  call from_history("d1",dz)
  call from_history("d2",dx)
  call from_history("o1",z0)
  call from_history("o2",x0)
  
  write(0,*) "nz,nx",nz,nx

  call from_param("npad",npad,20)  ! Padding zone

  call from_param("T",T)           ! Total recording time
  call from_param("freesurf",freesurf,.true.) ! Free (top) surface or not.
  
  
  !!! Medium Parameters
  allocate(vv(nz,nx)); vv = 0.
  allocate(rh(nz,nx)); rh = 0.
  call sreed("in",vv,4*size(vv))
  call sreed("in",rh,4*size(rh))
  call from_param("rhoc",rhoc,-1.)
  if (rhoc.ne.-1) then; rh = rhoc; end if

  !!! Determine timesteps
  vmax = 0.
  vmin = vv(1,1)
  do ix=1,nx;  do iz=1,nz
        vmax=max(vmax,vv(iz,ix))
        vmin=min(vmin,vv(iz,ix))
  end do    ;  end do
  dt_comp = min(dz,dx)/vmax/3.;
  nt_comp = T/dt_comp
  jt_verb = nint(1./dt_comp)
  
  ! *********************************************************************
  !  Source/Receicer Section
  ! *********************************************************************
  write(0,*) '----------------------------'
  write(0,*) 'Source/Receiver Setup'
  
  !! Receiver Output
  call from_param("nrt",nrt,-1)
  call from_param("nrz",nrz,1 )
  call from_param("nrx",nrx,1 )
  call from_param("frz",frz,nz/2+1)
  call from_param("frx",frx,nx/2+1)
  call from_param("jrz",jrz,1 )
  call from_param("jrx",jrx,1 )
  call from_param("rtype",rtype,1)
  nr = nrz*nrx
  if (exist_file("receivers_zxt")) then
    call from_aux("receivers_zxt","n2",nr,nr)
  end if
  allocate(receiver_zxt(nr,4));  receiver_zxt = 0;
  if (exist_file("receiver_zxt")) then
    allocate(receiver_zxt_in(nr,4)); receiver_zxt_in=0.
    call sreed("receiver_zxt",receiver_zxt_in,4*size(receiver_zxt_in))
    do ir=1,nr
      receiver_zxt(ir,:) = (/ 1+nint((receiver_zxt_in(ir,1)-z0)/dz), &
                              1+nint((receiver_zxt_in(ir,2)-x0)/dx), &
                              1+nint( receiver_zxt_in(ir,3)/dt_comp), int(receiver_zxt_in(ir,4)) /)
    end do
    deallocate(receiver_zxt_in)
  else
    do irx=1,nrx
    do irz=1,nrz
      receiver_zxt((irx-1)*nrz+irz,:) = (/ 1+frz+(irz-1)*jrz,1+frx+(irx-1)*jrx,1,rtype /)
    end do
    end do
  end if
  ! Allocate auxiliarry array for construction of Vx and Vz
  allocate(vprev(nr)); vprev=0.
  
  call from_param("dt_data",dt_data,dt_comp)
  jt_data = nint(dt_data/dt_comp)
  dt_data = dt_comp*jt_data      ! After rounding
  if (nrt.eq.-1) then; nrt = int(T/dt_data); end if
  nt_data = nrt
  allocate(data(nt_data,nr)); data=0.;
  
  ! Create output header for data
  call to_history("n1",nt_data)
  call to_history("n2",nr)
  call to_history("n3",1 )
  
  call to_history("d1",dt_data)
  call to_history("d2",dx*jrx)
  call to_history("d3",1.)
  
  call to_history("o1",0.)
  call to_history("o2",x0+frx*dx*jrx)
  call to_history("o3",0.)
  
  
  !! Snapshot output
  if (exist_file("snap")) then
    snap = .true.
    call from_param("dt_snap",dt_snap,dt_comp)
    jt_snap = nint(dt_snap/dt_comp)
    dt_snap = dt_comp*jt_snap      ! After rounding
    nt_snap = nint(T/dt_snap)  

    ! Create output header for snapshots
    call to_history("n1",nz,"snap")
    call to_history("n2",nx,"snap")
    call to_history("n3",nt_snap,"snap")
    
    call to_history("d1",dz,"snap")
    call to_history("d2",dx,"snap")
    call to_history("d3",dt_snap,"snap")
    
    call to_history("o1",z0,"snap")
    call to_history("o2",x0,"snap")
    call to_history("o3",0.,"snap")

    allocate(usnap (nz,nx)); usnap=0.;
  else
    snap = .false.
  end if
  
  
  !! Sources Input
  call from_param("nst",nst,128 )
  call from_param("nsz",nsz,1 )
  call from_param("nsx",nsx,1 )
  call from_param("fsz",fsz,nz/2+1)
  call from_param("fsx",fsx,nx/2+1)
  call from_param("jsz",jsz,1 )
  call from_param("jsx",jsx,1 )
  call from_param("stype",stype,1)
  ns = nsz*nsx
  if (exist_file("sources_zxt")) then
    call from_aux("sources_zxt","n1",ns,ns)
  end if
  allocate(sources_zxt(ns,4));  sources_zxt = 0.
  if (exist_file("sources_zxt")) then
    allocate(sources_zxt_in(ns,4)); sources_zxt_in = 0.
    call sreed("sources_zxt",sources_zxt_in,4*size(sources_zxt_in))
    do is=1,ns
      sources_zxt(is,:) = (/ 1+nint((sources_zxt_in(is,1)-z0)/dz), &
                             1+nint((sources_zxt_in(is,2)-x0)/dx), &
                             1+nint( sources_zxt_in(is,3)/dt_comp), int(sources_zxt_in(is,4)) /)
    end do
    deallocate(sources_zxt_in)
  else
    do isx=1,nsx
    do isz=1,nsz
      sources_zxt((isx-1)*nsz+isz,:) = (/ 1+fsz+(isz-1)*jsz,1+fsx+(isx-1)*jsx,1,stype /)
    end do
    end do
  end if
  !! Sources Function Input (or construction)
  allocate(sources(-nst:nst,ns));  sources = 0.
  call from_param("f0",f0,25.); w0 = 2.*pi*f0
  call from_param("ssquare",ssquare,.false.)
  call from_param("smth",smth,0) ! Source smoothing over smth gridpoints
  smth_fact = 0.   ! Weighting factor because of source smoothing
  do ix=-smth,smth
  do iz=-smth,smth
      rr        = sqrt(float((ix)**2+(iz)**2))
      smth_fact = smth_fact + .5*(cos(rr/3.*pi)+1.)
  end do
  end do
  do is=1,ns
    call sourcericker(sources(-nst:nst,is),2*nst,dt_comp,w0,1,ssquare)
    sources(-nst:nst,is) = sources(-nst:nst,is)*dt_comp*dt_comp*vv(sources_zxt(is,1),sources_zxt(is,2))*vv(sources_zxt(is,1),sources_zxt(is,2))/(dx*dz)/smth_fact
  end do
  ! Normalize the source function according to the update with dt**2
  
  call sep_close()

 ! write(0,*) 'sources_zxt(:,1)'
 ! write(0,*)  sources_zxt(:,1)
!  write(0,*) 'sources_zxt(:,2)'
!  write(0,*)  sources_zxt(:,2)
!  write(0,*) 'sources_zxt(:,3)'
!  write(0,*)  sources_zxt(:,3)
  
  !write(0,*) 'receiver_zxt(:,1)'
  !write(0,*)  receiver_zxt(:,1)
  !write(0,*) 'receiver_zxt(:,2)'
  !write(0,*)  receiver_zxt(:,2)
  !write(0,*) 'receiver_zxt(:,3)'
  !write(0,*)  receiver_zxt(:,3)
  
  
  ! *********************************************************************
  !  Setup Finite Differences
  ! *********************************************************************
  write(0,*) '----------------------------'
  write(0,*) 'Setup Finite Differences'
  
  !allocate space for the velocity model, the absorbing boundaries and the wavefields
  allocate(u(-4:nz+5,-4:nx+5,3)); u = 0.
  allocate(a0 (nz,nx));  a0   = 0.
  allocate(a1x(nz,nx));  a1x  = 0.
  allocate(a2x(nz,nx));  a2x  = 0.
  allocate(a3x(nz,nx));  a3x  = 0.
  allocate(a4x(nz,nx));  a4x  = 0.
  allocate(a5x(nz,nx));  a5x  = 0.
  allocate(a1z(nz,nx));  a1z  = 0.
  allocate(a2z(nz,nx));  a2z  = 0.
  allocate(a3z(nz,nx));  a3z  = 0.
  allocate(a4z(nz,nx));  a4z  = 0.
  allocate(a5z(nz,nx));  a5z  = 0.
  allocate(mult(nz,nx)); mult = 0.

  ! call the setup routine to build a propagator for the main grid
  call setupfd(vv,mult,a0,a1x,a2x,a3x,a4x,a5x,a1z,a2z,a3z,a4z,a5z,nz,nx,npad,dt_comp,dx,dz,freesurf)
  
  
  call stop_timer_num (tm2)
  call start_timer_num(tm3)
  ! *********************************************************************
  !  Loop over the total time
  ! *********************************************************************
  write(0,*) '----------------------------'
  write(0,*) 'Loop over the total time', nt_comp

  it=-nst+2
  id=1
  in=1

  do while (it.le.nt_comp)
    
    !!! do the global update thing   

    if (freesurf) then
       !! anti-symetrize the  p-wave potential above the free surface
       do ix=1,nx
             u(-1,ix,2) =  0. 
             u(-2,ix,2) =  0.
             u(-3,ix,2) =  0.
             u(-4,ix,2) =  0.
             u( 0,ix,2) =  0.
       end do
    end if

    !!! Calculate the Laplacian of the P wave field at the current time step
    do ix=1,nx
       do iz=1,nz
           u(iz,ix,3)=a0 (iz,ix)* u(iz,ix  ,2)              +           &
                      a1x(iz,ix)*(u(iz,ix+1,2)+u(iz,ix-1,2))+           &
                      a2x(iz,ix)*(u(iz,ix+2,2)+u(iz,ix-2,2))+           &
                      a3x(iz,ix)*(u(iz,ix+3,2)+u(iz,ix-3,2))+           &
                      a4x(iz,ix)*(u(iz,ix+4,2)+u(iz,ix-4,2))+           &
                      a5x(iz,ix)*(u(iz,ix+5,2)+u(iz,ix-5,2))+           &
                      a1z(iz,ix)*(u(iz+1,ix,2)+u(iz-1,ix,2))+           &
                      a2z(iz,ix)*(u(iz+2,ix,2)+u(iz-2,ix,2))+           &
                      a3z(iz,ix)*(u(iz+3,ix,2)+u(iz-3,ix,2))+           &
                      a4z(iz,ix)*(u(iz+4,ix,2)+u(iz-4,ix,2))+           &
                      a5z(iz,ix)*(u(iz+5,ix,2)+u(iz-5,ix,2))
       end do
    end do

    ! 'Add' the time derivative
    do ix=1,nx
       do iz=1,nz
          u(iz,ix,3) = u(iz,ix,3) + 2.*u(iz,ix,2) - u(iz,ix,1)
       end do
    end do
    !! Adjust the wave field in the absorbing boundaries using the
    !! Israeli-Orszag (J. Comp Phys 1981) method to remove waves coming
    !! back from the boundaries.    
    !  top sponge
    do iz=1,npad
       do ix=1,nx
          u(iz,ix,3)=u(iz,ix,3)-mult(iz,ix)*(u(iz,ix,3)-u(iz,ix,2)+vv(1,ix)*dt_comp/dz*(u(iz-1,ix,2)-u(iz,ix,2)))
       end do
    end do
    !  bottom sponge
    do ix=1,nx
       do iz=nz-npad,nz
         u(iz,ix,3)=u(iz,ix,3)-mult(iz,ix)*(u(iz,ix,3)-u(iz,ix,2)+vv(nz,ix)*dt_comp/dz*(u(iz+1,ix,2)-u(iz,ix,2)))
       end do
    end do
    ! right side sponge
    do ix=nx-npad,nx
       do iz=1,nz
          u(iz,ix,3)=u(iz,ix,3)-mult(iz,ix)*(u(iz,ix,3)-u(iz,ix,2)+vv(iz,nx)*dt_comp/dx*(u(iz,ix+1,2)-u(iz,ix,2)))
       end do
    end do
    ! left side sponge
    do ix=1,npad
       do iz=1,nz
          u(iz,ix,3)=u(iz,ix,3)-mult(iz,ix)*(u(iz,ix,3)-u(iz,ix,2)+vv(iz,1)*dt_comp/dx*(u(iz,ix-1,2)-u(iz,ix,2)))
       end do
    end do


    !! Add the source wavelet at the source location: The source is
    !! introduced over a small region and not at just one point to prevent
    !! energy at nyquist wavelength from entering the calculation
    do is=1,ns
       if(it-sources_zxt(is,3)+1.le.nst.and.it-sources_zxt(is,3)-1.ge.-nst)then
         ! write(0,*) "Inject Source:",is

         ! Classic right-hand side source
         if      (sources_zxt(is,4).eq.0) then
        !   if (it-sources_zxt(is,3).eq.0) write(0,*) "Inject Source:", is,sources(it-sources_zxt(is,3),is)
           do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
              do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
                 rr=sqrt(float((ix-sources_zxt(is,2))**2+(iz-sources_zxt(is,1))**2))
                 u(iz,ix,3)=u(iz,ix,3)+sources(it-sources_zxt(is,3),is)!*.5*(cos(rr/3*pi)+1.)
              end do
           end do
         
         ! Impulsive volume injection-rate source type ( wavelet dt already done by sourcericker() )
         else if (sources_zxt(is,4).eq.1) then

           do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
              do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
                 rr=sqrt(float((ix-sources_zxt(is,2))**2+(iz-sources_zxt(is,1))**2))
                 ! +
                 u(iz,ix,3)=u(iz,ix,3)+rh(iz,ix)/(2*dt_comp)*sources(it-sources_zxt(is,3)+1,is)*.5*(cos(rr/3*pi)+1.)
                 ! -
                 u(iz,ix,3)=u(iz,ix,3)-rh(iz,ix)/(2*dt_comp)*sources(it-sources_zxt(is,3)-1,is)*.5*(cos(rr/3*pi)+1.)
              end do
           end do

         ! Impulsive force source source type (x-direction)
         else if (sources_zxt(is,4).eq.2) then
           do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
              do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
                 ! +
                 rr=sqrt(float((ix+1-sources_zxt(is,2))**2+(iz-sources_zxt(is,1))**2))
                 u(iz,ix,3)=u(iz,ix,3)-1./(2*dx)*sources(it-sources_zxt(is,3),is)*.5*(cos(rr/3*pi)+1.)
                 ! -
                 rr=sqrt(float((ix-1-sources_zxt(is,2))**2+(iz-sources_zxt(is,1))**2))
                 u(iz,ix,3)=u(iz,ix,3)+1./(2*dx)*sources(it-sources_zxt(is,3),is)*.5*(cos(rr/3*pi)+1.)
              end do
           end do

         ! Impulsive force source source type (z-direction)
         else if (sources_zxt(is,4).eq.3) then
           do ix=sources_zxt(is,2)-smth,sources_zxt(is,2)+smth
              do iz=sources_zxt(is,1)-smth,sources_zxt(is,1)+smth
                 ! +
                 rr=sqrt(float((ix-sources_zxt(is,2))**2+(iz+1-sources_zxt(is,1))**2))
                 u(iz,ix,3)=u(iz,ix,3)-1./(2*dz)*sources(it-sources_zxt(is,3),is)*.5*(cos(rr/3*pi)+1.)
                 ! -
                 rr=sqrt(float((ix-sources_zxt(is,2))**2+(iz-1-sources_zxt(is,1))**2))
                 u(iz,ix,3)=u(iz,ix,3)+1./(2*dz)*sources(it-sources_zxt(is,3),is)*.5*(cos(rr/3*pi)+1.)
              end do
           end do
         end if

       end if
    end do
    


    !! Save the current wavefield at the geophones as our seismogram: Save the
    !! locations that were requested only.
    if ( mod(it,jt_data).eq.0 .and. it.ge.0 .and. id.le.nt_data ) then
       do ir=1,nr
          
          ! Record Pressure
          if      (receiver_zxt(ir,4).eq.1) then
            data(id,ir) =               u(receiver_zxt(ir,1)  ,receiver_zxt(ir,2)  ,3)

          ! Record Vx
          else if (receiver_zxt(ir,4).eq.2) then
            data(id,ir) = vprev(ir) - ( u(receiver_zxt(ir,1)  ,receiver_zxt(ir,2)+1,3) - u(receiver_zxt(ir,1)  ,receiver_zxt(ir,2)-1,3) )/(rh(iz,ix)*2*dx)*dt_comp
            vprev(  ir) = data(id,ir)
          
          ! Record Vz
          else if (receiver_zxt(ir,4).eq.3) then
            data(id,ir) = vprev(ir) - ( u(receiver_zxt(ir,1)+1,receiver_zxt(ir,2)  ,3) - u(receiver_zxt(ir,1)-1,receiver_zxt(ir,2)  ,3) )/(rh(iz,ix)*2*dx)*dt_comp
            vprev(  ir) = data(id,ir)

          end if
          
      end do
      id=id+1
    else if (id.le.nt_data) then
       do ir=1,nr
        
        ! Record Vx
        if      (receiver_zxt(ir,4).eq.2) then
          vprev(  ir) = vprev(ir) - ( u(receiver_zxt(ir,1)  ,receiver_zxt(ir,2)+1,3) - u(receiver_zxt(ir,1)  ,receiver_zxt(ir,2)-1,3) )/(rh(iz,ix)*2*dx)*dt_comp
          
        ! Record Vz
        else if (receiver_zxt(ir,4).eq.3) then
          vprev(  ir) = vprev(ir) - ( u(receiver_zxt(ir,1)+1,receiver_zxt(ir,2)  ,3) - u(receiver_zxt(ir,1)-1,receiver_zxt(ir,2)  ,3) )/(rh(iz,ix)*2*dx)*dt_comp

        end if

      end do
    end if
    

    !! Now output the desired snapshots if they were requested  
    !!   they are calculated at only every jt_snap time steps
    if ( snap ) then
      if ( mod(it,jt_snap).eq.0 .and. it.ge.0 .and. in.le.nt_snap ) then
           do ix=1,nx
           do iz=1,nz
              usnap(iz,ix)=u(iz,ix,3)
           end do
           end do
        in=in+1
        call srite("snap",usnap,4*nx*nz)
      end if
    end if



 !   !!! CYCLE
    
    ! Update  all the variables in time
    do ix=1,nx
        do iz=1,nz
           u(iz,ix,1)=u(iz,ix,2)
           u(iz,ix,2)=u(iz,ix,3)
        end do
    end do
    !!! Update the 1-way wave equation regions

    ! Update the bottom operator pad
    do iz=nz+1,nz+5
       do ix=1,nx
         delp=1.2*vv(nz,ix)*dt_comp/dz
         u(iz,ix,2)=1./(1.+delp)*((1-delp)*(u(iz,ix,1)-u(iz-1,ix,2)))+u(iz-1,ix,1)
       end do
    end do
    do iz=nz+1,nz+5
       do ix=1,nx
          u(iz,ix,1)=u(iz,ix,2)
       end do
    end do

    if (.not. freesurf) then
       ! Update the top side operator pad
       do iz=0,-4,-1
          do ix=1,nx
             delp=1.2*vv(1,ix)*dt_comp/dz
             u(iz,ix,2)=1./(1.+delp)*((1-delp)*(u(iz,ix,1)-u(iz+1,ix,2)))+u(iz+1,ix,1)
          end do
       end do
       do iz=0,-4,-1
          do ix=1,nz
             u(iz,ix,1)=u(iz,ix,2)
          end do
       end do
    end if

    ! Update the right side operator pad
    do ix=nx+1,nx+5
       do iz=1,nz
          delp=1.2*vv(iz,nx)*dt_comp/dx
          u(iz,ix,2)=1./(1.+delp)*((1-delp)*(u(iz,ix,1)-u(iz,ix-1,2)))+u(iz,ix-1,1)
       end do
    end do
    do ix=nx+1,nx+5
       do iz=1,nz
          u(iz,ix,1)=u(iz,ix,2)
       end do
    end do

    ! Update the left side operator pad
    do ix=0,-4,-1
       do iz=1,nz
          delp=1.2*vv(iz,1)*dt_comp/dx
          u(iz,ix,2)=1./(1.+delp)*((1-delp)*(u(iz,ix,1)-u(iz,ix+1,2)))+u(iz,ix+1,1)
       end do
    end do
    do ix=0,-4,-1
       do iz=1,nz
          u(iz,ix,1)=u(iz,ix,2)
       end do
    end do

    ! Time step verbose
    if (mod(it-1,jt_verb).eq.0.) then
      write(0,*) "Step:",nint(it*dt_comp),'seconds'
    end if
    
    ! Update time counter
    it = it+1
    
  end do
  ! *********************************************************************
  !  End of Loop over the total time
  ! *********************************************************************
  
  
  ! Write out data and snaps
  call srite("out" ,data ,4*size(data ))
 ! if ( snap ) then
 !   call srite("snap",usnap,4*size(usnap))
 ! end if

  ! Final
  write(0,*) '----------------------------'
  call stop_timer_num (tm3)
  call stop_timer_num (tm1)
  call print_timers()


  deallocate(vv )
  deallocate(rh )
  deallocate(u  )
  deallocate(a0 )
  deallocate(a1x)
  deallocate(a2x)
  deallocate(a3x)
  deallocate(a4x)
  deallocate(a5x)
  deallocate(a1z)
  deallocate(a2z)
  deallocate(a3z)
  deallocate(a4z)
  deallocate(a5z)
  deallocate(mult)
  deallocate(receiver_zxt)
  deallocate(sources_zxt)
  deallocate(data)
  deallocate(sources)
  if ( snap ) then
    deallocate(usnap)
  end if
  deallocate(vprev)
  
  write(0,*) '----------------------------'
  call exit(0)

end program












  subroutine setupfd(v,mult,a0,a1x,a2x,a3x,a4x,a5x,a1z,a2z,a3z,a4z,a5z,      &
                     nz,nx,npad,dt,dx,dz,freesurf)

    implicit none
    integer :: nz,nx,iz,ix,npad
    real    :: dt,dx,dz
    real    :: a0(nz,nx),a1x(nz,nx),a2x(nz,nx),a3x(nz,nx),a4x(nz,nx),a5x(nz,nx)
    real    ::           a1z(nz,nx),a2z(nz,nx),a3z(nz,nx),a4z(nz,nx),a5z(nz,nx)
    real    :: v(nz,nx),mult(nz,nx)
    logical :: freesurf

    do ix=1,nx
      do iz=1,nz

        a1x(iz,ix)=3.333333333*v(iz,ix)*v(iz,ix)*dt*dt/(2*dx*dx)
        a1z(iz,ix)=3.333333333*v(iz,ix)*v(iz,ix)*dt*dt/(2*dz*dz)

        a2x(iz,ix)=-.4761904762*v(iz,ix)*v(iz,ix)*dt*dt/(2*dx*dx)
        a2z(iz,ix)=-.4761904762*v(iz,ix)*v(iz,ix)*dt*dt/(2*dz*dz)

        a3x(iz,ix)=.0793650794*v(iz,ix)*v(iz,ix)*dt*dt/(2*dx*dx)
        a3z(iz,ix)=.0793650794*v(iz,ix)*v(iz,ix)*dt*dt/(2*dz*dz)

        a4x(iz,ix)=-.0099206349*v(iz,ix)*v(iz,ix)*dt*dt/(2*dx*dx)
        a4z(iz,ix)=-.0099206349*v(iz,ix)*v(iz,ix)*dt*dt/(2*dz*dz)

        a5x(iz,ix)=.0006349206*v(iz,ix)*v(iz,ix)*dt*dt/(2*dx*dx)
        a5z(iz,ix)=.0006349206*v(iz,ix)*v(iz,ix)*dt*dt/(2*dz*dz)

        a0(iz,ix)=          (a5x(iz,ix)+a5z(iz,ix))
        a0(iz,ix)=a0(iz,ix)+(a4x(iz,ix)+a4z(iz,ix))
        a0(iz,ix)=a0(iz,ix)+(a3x(iz,ix)+a3z(iz,ix))
        a0(iz,ix)=a0(iz,ix)+(a2x(iz,ix)+a2z(iz,ix))   ! done this crazy way to help prevent loss of precision
        a0(iz,ix)=a0(iz,ix)+(a1x(iz,ix)+a1z(iz,ix))
        a0(iz,ix)= -2.*a0(iz,ix)

      end do
    end do

    if (.not. freesurf) then
       ! top sponge coeffs
       do ix=1,nz
          do iz=1,npad
             mult(iz,ix)=float((npad-iz)+1)/float(npad)*.05*((v(iz,ix)*dt/dz)/.333)
          end do
       end do
    end if

    ! right sponge coeffs
    do ix=nx-npad,nx
       do iz=1,nz
          mult(iz,ix)=float(ix-(nx-npad)+1)/float(npad)*.05*((v(iz,ix)*dt/dx)/.333)
       end do
    end do

    ! left sponge coeffs
    do ix=1,npad
       do iz=1,nz
          mult(iz,ix)=float((npad-ix)+1)/float(npad)*.05*((v(iz,ix)*dt/dx)/.333)
       end do
    end do

    ! bottom sponge coeffs
    do ix=1,nx
       do iz=nz-npad,nz
          mult(iz,ix)=float(iz-(nz-npad)+1)/float(npad)*.05*((v(iz,ix)*dt/dz)/.333)
       end do
    end do

    return

  end subroutine setupfd


