module reg_laplac_mod
  ! Laplacian operator in 2-D
  
  implicit none

  double precision,private,  dimension (:), allocatable   :: reg_tmp
  integer,private                             :: nx,ny
  real,private                                :: dx,dy
  
  private
  public :: reg_laplac_init, reg_laplac_close, reg_laplac_lop, reg_laplac_op
  
  contains
  
  subroutine reg_laplac_init( nx_in,ny_in,dx_in,dy_in)
    integer nx_in, ny_in
    real    dx_in, dy_in
    call reg_laplac_close()
    nx = nx_in
    ny = ny_in
    dx = dx_in
    dy = dy_in
    allocate(reg_tmp(nx*ny*2)) 
  end subroutine
  
  subroutine reg_laplac_close()
    if (allocated(reg_tmp)) deallocate( reg_tmp )
  end subroutine
  
  
  
  ! Generic operator for SEP solvers
  function reg_laplac_lop(adj,add,m,d) result(stat)
    integer             :: stat
    logical,intent(in)  :: adj , add
    double precision,intent(inout)  :: m(:), d(:)
    call adj_null(adj,add,m,d)
    if (.not.adj) call reg_laplac_for(m,d)
    if (     adj) call reg_laplac_adj(m,d)
    stat=0
  end function

  subroutine reg_laplac_op(adj,add,m,d)
    integer             :: stat
    logical,intent(in)  :: adj , add
    double precision,intent(inout)  :: m(:), d(:)
    call adj_null(adj,add,m,d)
    if (.not.adj) call reg_laplac_for(m,d)
    if (     adj) call reg_laplac_adj(m,d)
  end subroutine
  



  ! Forward operator
  subroutine reg_laplac_for( m, d)
    double precision,intent(inout)  :: m(:),d(:)
    call igrad2_lop ( .false., .false., m, reg_tmp)     ! tmp = grad y
    call igrad2_lop (  .true.,  .true., d, reg_tmp)     ! x   = x + grad' tmp
  end subroutine

  ! Adjoint operator
  subroutine reg_laplac_adj( m, d)
    double precision,intent(inout)  :: m(:),d(:)
    call igrad2_lop ( .false., .false., d, reg_tmp)     ! tmp = grad x
    call igrad2_lop (  .true.,  .true., m, reg_tmp)     ! y   = y + grad' tmp
  end subroutine


  ! Gradient operators
  subroutine igrad2_lop(adj,add,m,d)
    logical,intent(in)    :: adj,add
    double precision,  intent(inout) ::   m(nx,ny),d(nx,ny,2)
    integer               :: ii,jj
    call adjnull(adj,add,m,d)
    
    if( adj ) then
      do ii= 1, nx-1
        do jj= 1, ny-1
	   m(ii+1,jj  ) = m(ii+1,jj  ) + d(ii,jj,1) 
           m(ii  ,jj  ) = m(ii  ,jj  ) - d(ii,jj,1) 
           m(ii  ,jj+1) = m(ii  ,jj+1) + d(ii,jj,2) 
           m(ii  ,jj  ) = m(ii  ,jj  ) - d(ii,jj,2)
       end do
     end do
     m = m/dx/dy
   else
     do ii= 1, nx-1
       do jj= 1, ny-1
         d(ii  ,jj,1) = d(ii  ,jj,1) + ( m(ii+1,jj) - m(ii,jj) ) 
         d(ii  ,jj,2) = d(ii  ,jj,2) + ( m(ii,jj+1) - m(ii,jj) ) 
       end do
     end do
     d = d/dx/dy
   end if

  end subroutine
  
  
  subroutine     adj_null(adj,add,x,y)
    logical,intent(in) :: adj , add
    double precision           :: x(:), y(:)
    if (.not.add) then
      if (     adj)  x = 0.
      if (.not.adj)  y = 0.
    end if
  end subroutine

  subroutine     adjnull(adj,add,x,y)
    logical,intent(in) :: adj , add
    double precision         :: x(:,:), y(:,:,:)
    if (.not.add) then
      if (     adj)  x = 0.
      if (.not.adj)  y = 0.
    end if
  end subroutine

end module 
